"""Application configuration and helper functions for check-dependencies."""

from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Iterable,
    Iterator,
    Sequence,
)

from check_dependencies.lib import Dependency, Module, Package, Packages
from check_dependencies.pyproject_toml import ConfigToml, PyProjectToml

if TYPE_CHECKING:
    import ast


logger = getLogger(__name__)


@dataclass()
class AppConfig:
    """Application config and helper functions."""

    known_extra: Collection[Package]
    known_missing: Collection[Module]
    provides: Packages
    dependencies: Collection[Package]
    pyproject_file: Path | None = None
    include_dev: bool = False
    verbose: bool = False
    show_all: bool = False

    @classmethod
    def from_cli_args(  # noqa: PLR0913 (factory method)
        cls,
        *,
        file_names: Sequence[str],
        known_extra: str = "",
        known_missing: str = "",
        provides: Iterable[str] = (),
        include_dev: bool = False,
        verbose: bool = False,
        show_all: bool = False,
        includes: Sequence[str] = (),
    ) -> AppConfig:
        """Create an AppConfig instance from CLI arguments.

        :file_names: List of file paths to analyze.
        :known_extra: Comma-separated list of known extra dependencies.
        :known_missing: Comma-separated list of known missing dependencies.
        :provides: Iterable of strings in the format "package=module" to specify
            provided modules.
        :include_dev: Whether to include development dependencies from pyproject.toml.
        :verbose: Whether to include detailed information in the output.
        :show_all: Whether to show all dependencies, including those that are OK.
        """
        provides_list: list[tuple[Package, Module]] = []
        for package_name, _, module in (
            map1.partition("=") for maps in provides for map1 in maps.split(",")
        ):
            if package_name.strip() and module.strip():
                provides_list.append((Package(package_name), Module(module)))

        if not file_names:
            file_names = ["."]

        src_cfg = PyProjectToml.for_paths(file_names, include_dev=include_dev)

        def with_includes(
            current_path: Path, paths: Iterable[Path | str], seen: set[Path]
        ) -> Iterable[ConfigToml]:
            for pth in paths:
                if (res_pth := (current_path / pth).resolve()) not in seen:
                    seen.add(res_pth)
                    cfg = ConfigToml.for_path(res_pth)
                    yield cfg
                    yield from with_includes(res_pth.parent, cfg.includes, seen)
                else:
                    logger.debug("Already parsed: %s", res_pth)

        seen: set[Path] = set()
        cfgs = [
            *with_includes(Path(), includes, seen),
            *with_includes(src_cfg.path.parent, src_cfg.includes, seen),
        ]

        def cfg_of(key: str) -> Iterable[Any]:
            yield from getattr(src_cfg, key)
            for cfg in cfgs:
                yield from getattr(cfg, key)

        return cls(
            include_dev=include_dev,
            verbose=verbose,
            show_all=show_all,
            known_extra=frozenset(
                pkg
                for pkg in (
                    *Package.set(known_extra.split(",")),
                    *cfg_of("known_extra"),
                )
                if pkg.canonical
            ),
            known_missing=frozenset(
                Module(module)
                for module in (*known_missing.split(","), *cfg_of("known_missing"))
                if module.strip()
            ),
            provides=Packages([*provides_list, *cfg_of("provides")]),
            dependencies=src_cfg.dependencies,
            pyproject_file=src_cfg.path,
        )

    def __post_init__(self) -> None:
        """Dataclass post init method to ensure sets are frozen."""
        self.known_extra = frozenset(filter(None, self.known_extra or ()))
        self.known_missing = frozenset(filter(None, self.known_missing or ()))

    def mk_src_formatter(
        self,
    ) -> Callable[[str, Dependency, Module, ast.AST | None], Iterator[str]]:
        """Formatter for missing or used dependencies."""
        cache: set[Module] = set()

        def src_cause_formatter(
            src_pth: str,
            cause: Dependency,
            module: Module,
            stmt: ast.AST | None,
        ) -> Iterator[str]:
            if self.verbose:
                if (
                    cause in (Dependency.NA, Dependency.FILE_ERROR, Dependency.UNKNOWN)
                    or self.show_all
                ):
                    location = (
                        f"{Path(src_pth).as_posix()}:{getattr(stmt, 'lineno', -1)}"
                    )
                    yield f"{cause.value}{cause.name} {location} {module.name}"
            elif cause == Dependency.FILE_ERROR:
                yield f"{cause.value} {src_pth}"
            elif module.raw:
                if cause in (Dependency.NA, Dependency.UNKNOWN) or self.show_all:
                    yield f"{cause.value} {module.name}"
            elif (pkg_ := module.main) not in cache:
                cache.add(pkg_)
                if cause == Dependency.NA or self.show_all:
                    yield f"{cause.value} {pkg_.name}"

        return src_cause_formatter

    def unused_fmt(self, module: str) -> Iterator[str]:
        """Formatter for unused but declared dependencies.

        :param module: The module name of the dependency.
        """
        name = Dependency.EXTRA.name if self.verbose else ""
        yield f"{Dependency.EXTRA.value}{name} {module}"



"""Main module for check_dependencies."""

from __future__ import annotations

import ast
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

from check_dependencies.builtin_module import BUILTINS
from check_dependencies.lib import Dependency, Module, Package, Packages

if TYPE_CHECKING:
    from collections.abc import Collection, Generator, Iterator

    from check_dependencies.app_config import AppConfig

logger = logging.getLogger("check_dependencies")
ERR_MISSING_DEPENDENCY = 2
ERR_EXTRA_DEPENDENCY = 4
ERR_NO_PYPROJECT = 8
ERR_FILE = 16


def yield_wrong_imports(
    file_names: Collection[str],
    app_cfg: AppConfig,
) -> Generator[str, None, int]:
    """Yield output lines of missing/unused imports.

    :param file_names: List of file paths to analyze.
    :param app_cfg: Application configuration containing dependencies and settings.
        If app_cfg.show_all is True, all imports are shown.
    """
    used_deps: set[Package] = set()
    src_fmt = app_cfg.mk_src_formatter()

    allowed_dependencies = {
        *app_cfg.dependencies,  # dependencies from pyproject.toml
        *app_cfg.known_extra,
        *Package.set(BUILTINS),
        *Package.set(x.name for x in app_cfg.known_missing),
    }
    provides = app_cfg.provides

    exit_status = 0

    for src_pth in _project_files(file_names):
        for cause, module, stmt in _missing_imports_iter(
            src_pth, dependencies=allowed_dependencies, provides=provides
        ):
            if cause not in (Dependency.OK, Dependency.FILE_ERROR, Dependency.UNKNOWN):
                exit_status |= ERR_MISSING_DEPENDENCY
            if not module.raw:
                used_deps |= provides.packages(module)
            yield from src_fmt(src_pth.as_posix(), cause, module, stmt)

    if superfluous_requirements := [
        msg
        for dep in sorted(
            set(app_cfg.dependencies).difference(used_deps, app_cfg.known_extra),
        )
        for msg in app_cfg.unused_fmt(str(dep))
    ]:
        exit_status |= ERR_EXTRA_DEPENDENCY
        if app_cfg.verbose:
            yield ""
            yield "### Dependencies in config file not used in application:"
            yield f"# Config file: {app_cfg.pyproject_file}"
        yield from superfluous_requirements
    return exit_status


def _project_files(file_names: Collection[str]) -> Iterator[Path]:
    """Collect all Python files in a list of files or directories.

    Ensure no file is visited more than once.

    :param file_names: List of file paths or directories to analyze.
    """
    visited = set()
    for p in map(Path, file_names):
        for p_sub in p.rglob("*.py") if p.is_dir() else [p]:
            # resolve to avoid duplicates from symlinks or different relative paths.
            # Symlink can be pointed outside the project directory.
            if (p_sub_resolved := p_sub.resolve()) not in visited:
                visited.add(p_sub_resolved)
                yield p_sub


def _missing_imports_iter(
    file: Path,
    dependencies: Collection[Package],
    provides: Packages,
) -> Iterator[tuple[Dependency, Module, ast.AST]]:
    """Find missing imports in a Python file.

    :param file: Python file to analyze
    :param dependencies: Declared dependencies from pyproject file
    :yields: Tuple of status, module and import statement
    """
    try:
        parsed = ast.parse(file.read_bytes(), filename=file.as_posix())
    except (SyntaxError, OSError, PermissionError, FileNotFoundError):
        logger.warning("Could not parse %s", file, exc_info=False)
        yield (
            Dependency.FILE_ERROR,
            Module(file.as_posix(), raw=True),
            ast.Raise(lineno=-1),
        )
        return
    for module, stmt in _imports_iter(parsed.body):
        if module.raw:
            yield Dependency.UNKNOWN, module, stmt
        else:
            pkg_ = provides.packages(module)
            status = Dependency.OK if pkg_.intersection(dependencies) else Dependency.NA
            yield status, module, stmt


def _imports_iter(body: list[ast.stmt]) -> Iterator[tuple[Module, ast.AST]]:
    """Yield all import statements from a body of code.

    :param body: List of AST statements to analyze.
    """
    for node in (node for stmt in body for node in ast.walk(stmt)):
        yield from _imports(node)
        yield from _import_builtin(node)


def _imports(stmt: ast.AST) -> Iterable[tuple[Module, ast.AST]]:
    """Yield all module names from an import statement."""
    if isinstance(stmt, ast.Import):
        for alias in stmt.names:
            yield Module(alias.name), stmt
    elif isinstance(stmt, ast.ImportFrom) and stmt.level == 0 and stmt.module:
        # level > 0 means relative import
        yield Module(stmt.module), stmt


def _import_builtin(stmt: ast.AST) -> Iterable[tuple[Module, ast.AST]]:
    if not isinstance(stmt, ast.Call):
        return

    if (id_ := _fq_call_name(stmt)) in ("__import__", "__builtins__.__import__"):
        if stmt.args:
            # __import__ is called with at least one argument, which is the module name
            arg = stmt.args[0]
        elif kw_name_arg := [kw.value for kw in stmt.keywords if kw.arg == "name"]:
            # __import__ is called with keyword __import(name=...)
            arg = kw_name_arg[0]
        else:
            # __import__ is called without arguments, which is invalid, so we ignore it.
            return

        if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
            yield Module(arg.value), stmt
        else:
            yield Module(f"{id_}(...)", raw=True), stmt


def _fq_call_name(stmt: ast.Call) -> str | None:
    """Get the fully qualified name of a function call, if possible."""
    func = stmt.func
    if isinstance(func, ast.Name):
        return func.id
    if isinstance(func, ast.Attribute) and isinstance(func.value, ast.Name):
        return f"{func.value.id}.{func.attr}"
    return None



"""A list of all builtin modules."""

from __future__ import annotations

import sys

BUILTINS: frozenset[str]
if sys.version_info >= (3, 10):
    BUILTINS = frozenset(sys.stdlib_module_names)  # pylint: disable=no-member
else:
    # Use the list of builtins from Python 3.9
    BUILTINS = frozenset(  # pragma: no cover
        [
            "__future__",
            "_abc",
            "_aix_support",
            "_ast",
            "_asyncio",
            "_bisect",
            "_blake2",
            "_bootlocale",
            "_bootsubprocess",
            "_bz2",
            "_codecs",
            "_codecs_cn",
            "_codecs_hk",
            "_codecs_iso2022",
            "_codecs_jp",
            "_codecs_kr",
            "_codecs_tw",
            "_collections",
            "_collections_abc",
            "_compat_pickle",
            "_compression",
            "_contextvars",
            "_crypt",
            "_csv",
            "_ctypes",
            "_ctypes_test",
            "_curses",
            "_curses_panel",
            "_datetime",
            "_dbm",
            "_decimal",
            "_distutils_hack",
            "_elementtree",
            "_functools",
            "_gdbm",
            "_hashlib",
            "_heapq",
            "_imp",
            "_io",
            "_json",
            "_locale",
            "_lsprof",
            "_lzma",
            "_markupbase",
            "_md5",
            "_multibytecodec",
            "_multiprocessing",
            "_opcode",
            "_operator",
            "_osx_support",
            "_peg_parser",
            "_pickle",
            "_posixshmem",
            "_posixsubprocess",
            "_py_abc",
            "_pydecimal",
            "_pyio",
            "_queue",
            "_random",
            "_scproxy",
            "_sha1",
            "_sha256",
            "_sha3",
            "_sha512",
            "_signal",
            "_sitebuiltins",
            "_socket",
            "_sqlite3",
            "_sre",
            "_ssl",
            "_stat",
            "_statistics",
            "_string",
            "_strptime",
            "_struct",
            "_symtable",
            "_sysconfigdata__darwin_darwin",
            "_testbuffer",
            "_testcapi",
            "_testimportmultiple",
            "_testinternalcapi",
            "_testmultiphase",
            "_thread",
            "_threading_local",
            "_tkinter",
            "_tracemalloc",
            "_uuid",
            "_virtualenv",
            "_warnings",
            "_weakref",
            "_weakrefset",
            "_xxsubinterpreters",
            "_xxtestfuzz",
            "_zoneinfo",
            "abc",
            "aifc",
            "antigravity",
            "argparse",
            "array",
            "ast",
            "asynchat",
            "asyncio",
            "asyncore",
            "atexit",
            "audioop",
            "base64",
            "bdb",
            "binascii",
            "binhex",
            "bisect",
            "builtins",
            "bz2",
            "cProfile",
            "calendar",
            "cgi",
            "cgitb",
            "chunk",
            "cmath",
            "cmd",
            "code",
            "codecs",
            "codeop",
            "collections",
            "colorsys",
            "compileall",
            "concurrent",
            "configparser",
            "contextlib",
            "contextvars",
            "copy",
            "copyreg",
            "crypt",
            "csv",
            "ctypes",
            "curses",
            "dataclasses",
            "datetime",
            "dbm",
            "decimal",
            "difflib",
            "dis",
            "distutils",
            "doctest",
            "email",
            "encodings",
            "ensurepip",
            "enum",
            "errno",
            "faulthandler",
            "fcntl",
            "filecmp",
            "fileinput",
            "fnmatch",
            "formatter",
            "fractions",
            "ftplib",
            "functools",
            "gc",
            "genericpath",
            "getopt",
            "getpass",
            "gettext",
            "glob",
            "graphlib",
            "grp",
            "gzip",
            "hashlib",
            "heapq",
            "hmac",
            "html",
            "http",
            "idlelib",
            "imaplib",
            "imghdr",
            "imp",
            "importlib",
            "inspect",
            "io",
            "ipaddress",
            "itertools",
            "json",
            "keyword",
            "lib2to3",
            "linecache",
            "locale",
            "logging",
            "lzma",
            "mailbox",
            "mailcap",
            "marshal",
            "math",
            "memray",
            "mimetypes",
            "mmap",
            "modulefinder",
            "multiprocessing",
            "netrc",
            "nis",
            "nntplib",
            "ntpath",
            "nturl2path",
            "numbers",
            "opcode",
            "operator",
            "optparse",
            "os",
            "parser",
            "pathlib",
            "pdb",
            "pickle",
            "pickletools",
            "pip",
            "pipes",
            "pkg_resources",
            "pkgutil",
            "platform",
            "plistlib",
            "poplib",
            "posix",
            "posixpath",
            "pprint",
            "profile",
            "pstats",
            "pty",
            "pwd",
            "py_compile",
            "pyclbr",
            "pydoc",
            "pydoc_data",
            "pyexpat",
            "queue",
            "quopri",
            "random",
            "re",
            "readline",
            "reprlib",
            "resource",
            "rlcompleter",
            "runpy",
            "sched",
            "secrets",
            "select",
            "selectors",
            "setuptools",
            "shelve",
            "shlex",
            "shutil",
            "signal",
            "site",
            "smtpd",
            "smtplib",
            "sndhdr",
            "socket",
            "socketserver",
            "sqlite3",
            "sre_compile",
            "sre_constants",
            "sre_parse",
            "ssl",
            "stat",
            "statistics",
            "string",
            "stringprep",
            "struct",
            "subprocess",
            "sunau",
            "symbol",
            "symtable",
            "sys",
            "sysconfig",
            "syslog",
            "tabnanny",
            "tarfile",
            "telnetlib",
            "tempfile",
            "termios",
            "test",
            "textwrap",
            "this",
            "threading",
            "time",
            "timeit",
            "tkinter",
            "token",
            "tokenize",
            "trace",
            "traceback",
            "tracemalloc",
            "tty",
            "turtle",
            "turtledemo",
            "types",
            "typing",
            "unicodedata",
            "unittest",
            "urllib",
            "uu",
            "uuid",
            "venv",
            "warnings",
            "wave",
            "weakref",
            "webbrowser",
            "wheel",
            "wsgiref",
            "xdrlib",
            "xml",
            "xmlrpc",
            "xxlimited",
            "xxsubtype",
            "zipapp",
            "zipfile",
            "zipimport",
            "zlib",
            "zoneinfo",
        ],
    )



"""Tests for the lib module."""

from __future__ import annotations

import ast

import pytest

from check_dependencies.app_config import AppConfig
from check_dependencies.lib import Dependency, Module, Package, Packages, _canonical


class TestModule:
    """Test suite for the Module."""

    def test__lt__(self) -> None:
        """Test comparison."""
        assert Module("a") < Module("x")
        assert Module("x", raw=False) < Module("a", raw=True)
        assert Module("a", raw=True) < Module("x", raw=True)

    def test__lt___notimplemented(self) -> None:
        """Test comparison against non-Module returns NotImplemented."""
        assert Module("foo").__lt__(0) is NotImplemented

    def test__eq__(self) -> None:
        """Test equality."""
        assert Module("foo") == Module("foo")
        assert Module("foo", raw=False) == Module("foo", raw=False)
        assert Module("foo", raw=True) == Module("foo", raw=True)
        assert Module("foo") != Module("foo", raw=True)
        assert Module("foo").__eq__(0) is NotImplemented

    @pytest.mark.parametrize(
        "module, expected",
        [
            (Module("foo"), "Module('foo')"),
            (Module("foo", raw=False), "Module('foo')"),
            (Module("foo", raw=True), "Module('foo', raw=True)"),
        ],
    )
    def test__repr__(self, module: Module, expected: str) -> None:
        """Test __repr__."""
        assert repr(module) == expected

    @pytest.mark.parametrize(
        "module, expected",
        [
            (Module("numpy.linalg").main, Module("numpy")),
            (Module("sklearn").main, Module("sklearn")),
            (Module("PIL.Image").main, Module("PIL")),
            (Module("foo.bar.baz", raw=True).main, Module("foo.bar.baz", raw=True)),
        ],
    )
    def test_main(self, module: Module, expected: Module) -> None:
        """Test the main property."""
        assert module.main == expected


class TestPackage:
    """Test suite for the Package value object."""

    @pytest.mark.parametrize(
        "left,right",
        [
            ("PyJWT", "pyjwt"),
            ("scikit-learn", "scikit_learn"),
            ("SciKit-Learn>=1.0", "scikit_learn"),
            ("Scikit-Learn>=1", "scikit-Learn==*"),
        ],
    )
    def test_equal_packages_share_canonical_and_hash(
        self, left: str, right: str
    ) -> None:
        """Equivalent package spellings compare and hash equally."""
        a = Package(left)
        b = Package(right)

        assert a == b
        assert a.canonical == b.canonical
        assert hash(a) == hash(b)

    @pytest.mark.parametrize(
        "left,right",
        [
            ("pytest", "pyyaml"),
            ("requests", "requestx"),
        ],
    )
    def test_different_packages_are_not_equal(self, left: str, right: str) -> None:
        """Different package names should not compare as equal."""
        assert Package(left) != Package(right)

    @pytest.mark.parametrize(
        "value,other",
        [
            ("PyJWT", "pyjwt"),
            ("scikit-learn", "scikit_learn"),
        ],
    )
    def test_package_equals_matching_string(self, value: str, other: str) -> None:
        """Package equality also works against canonical-equivalent strings."""
        assert Package(value) == other

    @pytest.mark.parametrize(
        "raw,expected_str,expected_bool",
        [
            ("PyJWT", "PyJWT", True),
            ("  ", "", False),
            ("", "", False),
        ],
    )
    def test_string_and_bool_behavior(
        self, raw: str, expected_str: str, expected_bool: bool
    ) -> None:
        """Original name is kept for display while truthiness uses canonical name."""
        package = Package(raw)
        assert str(package) == expected_str
        assert bool(package) is expected_bool

    def test_eq_not_implemented(self) -> None:
        """Test that we return NotImplemented for unknown comparison."""
        assert Package("foo").__eq__(0) is NotImplemented

    def test_gt_not_implemented(self) -> None:
        """Test that we return NotImplemented for unknown comparison."""
        assert Package("foo").__gt__(0) is NotImplemented

    def test_cmp_str(self) -> None:
        """Test comparison against a string."""
        assert Package("a==1.0.0") < "b"

    @pytest.mark.parametrize(
        "requirement, expected_module",
        [
            ("foo > 0", "foo"),
            ("scikit-learn", "scikit_learn"),
            ("SciKit-Learn >= 1.0", "scikit_learn"),
        ],
    )
    def test_modules_fallback_uses_canonical_name(
        self, requirement: str, expected_module: str
    ) -> None:
        """Fallback for unmapped packages should be a canonical module name."""
        packages = Packages([])  # no explicit mapping -> fallback path
        assert packages.modules(Package(requirement)) == {Module(expected_module)}

    def test___repr__(self) -> None:
        """Test __repr__."""
        assert repr(Package("foo")) == "Package('foo')"


class TestPackages:
    """Test suite for the Packages class, which manages package-module mappings."""

    def test_packages_multi_module_and_multi_package_mapping(self) -> None:
        """Packages supports many-to-many mappings.

        - one package can provide multiple modules
        - one module can be provided by multiple packages
        """
        pkg_a, pkg_b, pkg_c = (Package(f"pkg_{x}") for x in "abc")
        mod_common = Module("mod_common")
        mod_a_only = Module("mod_a_only")
        mod_b_only = Module("mod_b_only")
        mod_c_only = Module("mod_c_only")
        packages = Packages(
            [
                (pkg_a, mod_common),
                (pkg_a, mod_a_only),
                (pkg_b, mod_common),
                (pkg_b, mod_b_only),
                (pkg_c, mod_c_only),
            ]
        )

        # package -> modules (one package provides multiple modules)
        assert packages.modules(pkg_a) == {mod_common, mod_a_only}
        assert packages.modules(pkg_b) == {mod_common, mod_b_only}
        assert packages.modules(pkg_c) == {mod_c_only}

        # module -> packages (one module provided by multiple packages)
        assert packages.packages(mod_common) == {pkg_a, pkg_b}
        assert packages.packages(mod_a_only) == {pkg_a}
        assert packages.packages(mod_b_only) == {pkg_b}
        assert packages.packages(mod_c_only) == {pkg_c}


class TestNormalizePkg:
    """Test suite for the normalize_pkg helper."""

    @pytest.mark.parametrize(
        "name, expected",
        [
            ("pyjwt", "pyjwt"),
            ("PyJWT", "pyjwt"),
            ("scikit-learn", "scikit_learn"),
            ("scikit_learn", "scikit_learn"),
            ("SciKit-Learn", "scikit_learn"),
            ("Pillow", "pillow"),
            ("SciKit-Learn>=10.0", "scikit_learn"),
        ],
    )
    def test__canonical(self, name: str, expected: str) -> None:
        """_canonical lowercases and replaces hyphens with underscores."""
        assert _canonical(name) == expected


class TestMkSrcFormatter:
    """Test suite for the mk_src_formatter function."""

    @pytest.fixture
    def stmt(self) -> ast.stmt:
        """AST import statement fixture."""
        return ast.parse("import foo.bar").body[0]

    @pytest.mark.parametrize("verbose", [True, False])
    def test_no_show_all_on_status_ok(self, stmt: ast.stmt, verbose: bool) -> None:
        """If the import is expected, we do not show it."""
        cfg = AppConfig.from_cli_args(
            file_names=(), verbose=verbose, show_all=False, known_extra="foo"
        )
        fn = cfg.mk_src_formatter()
        assert not list(fn("src.py", Dependency.OK, Module("foo"), stmt))

    @pytest.mark.parametrize(
        "verbose, show_all, cause, expected",
        [
            (True, False, "!", "!NA src.py:1 foo"),
            (True, True, "!", "!NA src.py:1 foo"),
            (True, True, " ", " OK src.py:1 foo"),
            (False, False, "!", "! foo"),
            (False, True, "!", "! foo"),
            (False, True, " ", "  foo"),
        ],
    )
    def test(  # pylint: disable=too-many-arguments
        self,
        stmt: ast.stmt,
        verbose: bool,
        show_all: bool,
        cause: str,
        expected: str,
    ) -> None:
        """MkSrcFormatter generic tests."""
        cfg = AppConfig.from_cli_args(file_names=(), verbose=verbose, show_all=show_all)
        fn = cfg.mk_src_formatter()
        assert next(fn("src.py", Dependency(cause), Module("foo"), stmt)) == expected

    def test_cache(self, stmt: ast.stmt) -> None:
        """Test the cache mechanism for the formatter."""
        cfg = AppConfig.from_cli_args(file_names=(), verbose=False)
        fn = cfg.mk_src_formatter()
        assert list(fn("src.py", Dependency.NA, Module("foo"), stmt))
        assert not list(fn("src.py", Dependency.NA, Module("foo"), stmt))



"""Common fixtures and constants for tests."""

from pathlib import Path

import pytest

DATA = (Path(__file__).parent / "data").resolve()
SRC_MODULE = DATA / "src"
SRC = (SRC_MODULE / "src.py").as_posix()
SRC_UNICODE = (DATA / "src_unicode.py").as_posix()
PYPROJECT_UNICODE = DATA / "pyproject_unicode.toml"
PEP631 = DATA / "pyproject_pep631.toml"
PEP631_EXTRA = DATA / "pyproject_pep631_extra.toml"
POETRY = DATA / "pyproject_poetry.toml"
POETRY_EXTRA = DATA / "pyproject_poetry_extra.toml"
HATCH = DATA / "pyproject_hatch.toml"
HATCH_EXTRA = DATA / "pyproject_hatch_extra.toml"
UV_LEGACY = DATA / "pyproject_uv_legacy.toml"
UV_LEGACY_EXTRA = DATA / "pyproject_uv_legacy_extra.toml"
PYPROJECT_CFG = DATA / "pyproject_cfg.toml"
PYPROJECT_EMPTY = DATA / "pyproject_empty.toml"
PYPROJECT_PROVIDES = DATA / "pyproject_pep631_provides.toml"


@pytest.fixture(params=[PEP631, POETRY, HATCH, UV_LEGACY])
def pyproject(request: pytest.FixtureRequest) -> Path:
    """Fixture for testing pyproject.toml files."""
    return request.param


@pytest.fixture(params=[PEP631_EXTRA, POETRY_EXTRA, HATCH_EXTRA, UV_LEGACY_EXTRA])
def pyproject_extra(request: pytest.FixtureRequest) -> Path:
    """Fixture for testing extra requirements."""
    return request.param
